"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# Metadata for geedim supported Earth Engine collections
collection_info = {
"landsat7_c2_l2": {
    "start_date": "1999-01-01",
    "end_date": None,
    "ee_coll_name": "LANDSAT/LE07/C02/T1_L2",
    "bands": [
        {"id": "SR_B1", "name": "blue", "abbrev": "B", "bw_start": 0.45, "bw_end": 0.52, "res": 30},
        {"id": "SR_B2", "name": "green", "abbrev": "G", "bw_start": 0.52, "bw_end": 0.60, "res": 30},
        {"id": "SR_B3", "name": "red", "abbrev": "R", "bw_start": 0.63, "bw_end": 0.69, "res": 30},
        {"id": "SR_B4", "name": "near infrared", "abbrev": "NIR", "bw_start": 0.77, "bw_end": 0.90, "res": 30},
        {"id": "SR_B5", "name": "shortwave infrared 1", "abbrev": "SWIR1", "bw_start": 1.55, "bw_end": 1.75, "res": 30},
        {"id": "SR_B7", "name": "shortwave infrared 2", "abbrev": "SWIR2", "bw_start": 2.08, "bw_end": 2.35, "res": 30},
        {"id": "ST_B6", "name": "brightness temperature", "abbrev": "BT", "bw_start": 10.40, "bw_end": 12.50, "res": 30},
    ],
    "properties": [
        {"PROPERTY": "system:id", "ABBREV": "ID", "DESCRIPTION": "Earth Engine image id"},
        {"PROPERTY": "system:time_start", "ABBREV": "DATE", "DESCRIPTION": "Image capture date/time (UTC)"},
        {"PROPERTY": "VALID_PORTION", "ABBREV": "VALID", "DESCRIPTION": "Portion of cloud and shadow free pixels (%)"},
        {"PROPERTY": "AVG_SCORE", "ABBREV": "SCORE", "DESCRIPTION": "Average distance to cloud/shadow (m)"},
        {"PROPERTY": "GEOMETRIC_RMSE_MODEL", "ABBREV": "GRMSE", "DESCRIPTION": "Orthorectification RMSE (m)"},
        {"PROPERTY": "SUN_AZIMUTH", "ABBREV": "SAA", "DESCRIPTION": "Solar azimuth angle (deg)"},
        {"PROPERTY": "SUN_ELEVATION", "ABBREV": "SEA", "DESCRIPTION": "Solar elevation angle (deg)"}
    ]},
"landsat8_c2_l2": {
    "start_date": "2013-04-11",
    "end_date": None,
    "ee_coll_name": "LANDSAT/LC08/C02/T1_L2",
    "bands": [
        {"id": "SR_B1", "name": "ultra blue", "abbrev": "UB", "bw_start": 0.435, "bw_end": 0.451, "res": 30},
        {"id": "SR_B2", "name": "blue", "abbrev": "B", "bw_start": 0.452, "bw_end": 0.512, "res": 30},
        {"id": "SR_B3", "name": "green", "abbrev": "G", "bw_start": 0.533, "bw_end": 0.590, "res": 30},
        {"id": "SR_B4", "name": "red", "abbrev": "R", "bw_start": 0.636, "bw_end": 0.673, "res": 30},
        {"id": "SR_B5", "name": "near infrared", "abbrev": "NIR", "bw_start": 0.851, "bw_end": 0.879, "res": 30},
        {"id": "SR_B6", "name": "shortwave infrared 1", "abbrev": "SWIR1", "bw_start": 1.566, "bw_end": 1.651, "res": 30},
        {"id": "SR_B7", "name": "shortwave infrared 2", "abbrev": "SWIR2", "bw_start": 2.107, "bw_end": 2.294, "res": 30},
        {"id": "ST_B10", "name": "brightness temperature", "abbrev": "BT", "bw_start": 10.60, "bw_end": 11.19, "res": 30}
    ],
    "properties": [
        {"PROPERTY": "system:id", "ABBREV": "ID", "DESCRIPTION": "Earth Engine image id"},
        {"PROPERTY": "system:time_start", "ABBREV": "DATE", "DESCRIPTION": "Image capture date/time (UTC)"},
        {"PROPERTY": "VALID_PORTION", "ABBREV": "VALID", "DESCRIPTION": "Portion of cloud and shadow free pixels (%)"},
        {"PROPERTY": "AVG_SCORE", "ABBREV": "SCORE", "DESCRIPTION": "Average distance to cloud/shadow (m)"},
        {"PROPERTY": "GEOMETRIC_RMSE_MODEL", "ABBREV": "GRMSE", "DESCRIPTION": "Orthorectification RMSE (m)"},
        {"PROPERTY": "SUN_AZIMUTH", "ABBREV": "SAA", "DESCRIPTION": "Solar azimuth angle (deg)"},
        {"PROPERTY": "SUN_ELEVATION", "ABBREV": "SEA", "DESCRIPTION": "Solar elevation angle (deg)"}
    ]},
"sentinel2_toa": {
    "start_date": "2015-06-23",
    "end_date": None,
    "ee_coll_name": "COPERNICUS/S2",
    "bands": [
        {"id": "B1", "name": "aerosols", "abbrev": "UB", "bw_start": 0.421, "bw_end": 0.447, "res": 60},
        {"id": "B2", "name": "blue", "abbrev": "B", "bw_start": 0.439, "bw_end": 0.535, "res": 10},
        {"id": "B3", "name": "green", "abbrev": "G", "bw_start": 0.537, "bw_end": 0.582, "res": 10},
        {"id": "B4", "name": "red", "abbrev": "R", "bw_start": 0.646, "bw_end": 0.685, "res": 10},
        {"id": "B5", "name": "red edge 1", "abbrev": "RE1", "bw_start": 0.694, "bw_end": 0.714, "res": 20},
        {"id": "B6", "name": "red edge 2", "abbrev": "RE2", "bw_start": 0.731, "bw_end": 0.749, "res": 20},
        {"id": "B7", "name": "red edge 3", "abbrev": "RE3", "bw_start": 0.768, "bw_end": 0.796, "res": 20},
        {"id": "B8", "name": "near infrared 1", "abbrev": "NIR1", "bw_start": 0.767, "bw_end": 0.908, "res": 10},
        {"id": "B8A", "name": "red edge 4", "abbrev": "RE4", "bw_start": 0.848, "bw_end": 0.881, "res": 20},
        {"id": "B9", "name": "water vapour", "abbrev": "WV", "bw_start": 0.931, "bw_end": 0.958, "res": 60},
        {"id": "B11", "name": "shortwave infrared 1", "abbrev": "SWIR1", "bw_start": 1.539, "bw_end": 1.681, "res": 20},
        {"id": "B12", "name": "shortwave infrared 2", "abbrev": "SWIR2", "bw_start": 2.072, "bw_end": 2.312, "res": 20}
    ],
    "properties": [
        {"PROPERTY": "system:id", "ABBREV": "ID", "DESCRIPTION": "Earth Engine image id"},
        {"PROPERTY": "system:time_start", "ABBREV": "DATE", "DESCRIPTION": "Image capture date/time (UTC)"},
        {"PROPERTY": "VALID_PORTION", "ABBREV": "VALID", "DESCRIPTION": "Portion of cloud and shadow free pixels (%)"},
        {"PROPERTY": "AVG_SCORE", "ABBREV": "SCORE", "DESCRIPTION": "Average distance to cloud/shadow (m)"},
        {"PROPERTY": "RADIOMETRIC_QUALITY", "ABBREV": "RADQ", "DESCRIPTION": "Radiometric quality check"},
        {"PROPERTY": "GEOMETRIC_QUALITY", "ABBREV": "GEOMQ", "DESCRIPTION": "Geometric quality check"},
        {"PROPERTY": "MEAN_SOLAR_AZIMUTH_ANGLE", "ABBREV": "SAA", "DESCRIPTION": "Solar azimuth angle (deg)"},
        {"PROPERTY": "MEAN_SOLAR_ZENITH_ANGLE", "ABBREV": "SZA", "DESCRIPTION": "Solar zenith angle (deg)"},
        {"PROPERTY": "MEAN_INCIDENCE_AZIMUTH_ANGLE_B1", "ABBREV": "VAA", "DESCRIPTION": "View (B1) azimuth angle (deg)"},
        {"PROPERTY": "MEAN_INCIDENCE_ZENITH_ANGLE_B1", "ABBREV": "VZA", "DESCRIPTION": "View (B1) zenith angle (deg)"}
    ]},
"sentinel2_sr": {
    "start_date": "2017-03-28",
    "end_date": None,
    "ee_coll_name": "COPERNICUS/S2_SR",
    "bands": [
        {"id": "B1", "name": "aerosols", "abbrev": "UB", "bw_start": 0.421, "bw_end": 0.447, "res": 60},
        {"id": "B2", "name": "blue", "abbrev": "B", "bw_start": 0.439, "bw_end": 0.535, "res": 10},
        {"id": "B3", "name": "green", "abbrev": "G", "bw_start": 0.537, "bw_end": 0.582, "res": 10},
        {"id": "B4", "name": "red", "abbrev": "R", "bw_start": 0.646, "bw_end": 0.685, "res": 10},
        {"id": "B5", "name": "red edge 1", "abbrev": "RE1", "bw_start": 0.694, "bw_end": 0.714, "res": 20},
        {"id": "B6", "name": "red edge 2", "abbrev": "RE2", "bw_start": 0.731, "bw_end": 0.749, "res": 20},
        {"id": "B7", "name": "red edge 3", "abbrev": "RE3", "bw_start": 0.768, "bw_end": 0.796, "res": 20},
        {"id": "B8", "name": "near infrared 1", "abbrev": "NIR1", "bw_start": 0.767, "bw_end": 0.908, "res": 10},
        {"id": "B8A", "name": "red edge 4", "abbrev": "RE4", "bw_start": 0.848, "bw_end": 0.881, "res": 20},
        {"id": "B9", "name": "water vapour", "abbrev": "WV", "bw_start": 0.931, "bw_end": 0.958, "res": 60},
        {"id": "B11", "name": "shortwave infrared 1", "abbrev": "SWIR1", "bw_start": 1.539, "bw_end": 1.681, "res": 20},
        {"id": "B12", "name": "shortwave infrared 2", "abbrev": "SWIR2", "bw_start": 2.072, "bw_end": 2.312, "res": 20}
    ],
    "properties": [
        {"PROPERTY": "system:id", "ABBREV": "ID", "DESCRIPTION": "Earth Engine image id"},
        {"PROPERTY": "system:time_start", "ABBREV": "DATE", "DESCRIPTION": "Image capture date/time (UTC)"},
        {"PROPERTY": "VALID_PORTION", "ABBREV": "VALID", "DESCRIPTION": "Portion of cloud and shadow free pixels (%)"},
        {"PROPERTY": "AVG_SCORE", "ABBREV": "SCORE", "DESCRIPTION": "Average distance to cloud/shadow (m)"},
        {"PROPERTY": "RADIOMETRIC_QUALITY", "ABBREV": "RADQ", "DESCRIPTION": "Radiometric quality check"},
        {"PROPERTY": "GEOMETRIC_QUALITY", "ABBREV": "GEOMQ", "DESCRIPTION": "Geometric quality check"},
        {"PROPERTY": "MEAN_SOLAR_AZIMUTH_ANGLE", "ABBREV": "SAA", "DESCRIPTION": "Solar azimuth angle (deg)"},
        {"PROPERTY": "MEAN_SOLAR_ZENITH_ANGLE", "ABBREV": "SZA", "DESCRIPTION": "Solar zenith angle (deg)"},
        {"PROPERTY": "MEAN_INCIDENCE_AZIMUTH_ANGLE_B1", "ABBREV": "VAA", "DESCRIPTION": "View (B1) azimuth angle (deg)"},
        {"PROPERTY": "MEAN_INCIDENCE_ZENITH_ANGLE_B1", "ABBREV": "VZA", "DESCRIPTION": "View (B1) zenith angle (deg)"}
    ]},
"modis_nbar": {
    "start_date": "2000-02-18",
    "end_date": None,
    "ee_coll_name": "MODIS/006/MCD43A4",
    "bands": [
        {"id": "Nadir_Reflectance_Band1", "name": "red", "abbrev": "R", "bw_start": 0.620, "bw_end": 0.670, "res": 500},
        {"id": "Nadir_Reflectance_Band2", "name": "near infrared", "abbrev": "NIR", "bw_start": 0.841, "bw_end": 0.876, "res": 500},
        {"id": "Nadir_Reflectance_Band3", "name": "blue", "abbrev": "B", "bw_start": 0.459, "bw_end": 0.479, "res": 500},
        {"id": "Nadir_Reflectance_Band4", "name": "green", "abbrev": "G", "bw_start": 0.545, "bw_end": 0.565, "res": 500},
        {"id": "Nadir_Reflectance_Band5", "name": "shortwave infrared 1", "abbrev": "SWIR1", "bw_start": 1.230, "bw_end": 1.250, "res": 500},
        {"id": "Nadir_Reflectance_Band6", "name": "shortwave infrared 2", "abbrev": "SWIR2", "bw_start": 1.628, "bw_end": 1.652, "res": 500},
        {"id": "Nadir_Reflectance_Band7", "name": "shortwave infrared 3", "abbrev": "SWIR3", "bw_start": 2.105, "bw_end": 2.155, "res": 500}
    ],
    "properties": [
        {"PROPERTY": "system:id", "ABBREV": "ID", "DESCRIPTION": "Earth Engine image id"},
        {"PROPERTY": "system:time_start", "ABBREV": "DATE", "DESCRIPTION": "Image capture date/time (UTC)"}
    ]}
}

# Dict to convert from geedim to Earth Engine collection names
gd_to_ee = dict([(k, v['ee_coll_name']) for k, v in collection_info.items()])

# Dict to convert from Earth Engine to geedim collection names
ee_to_gd = dict([(v['ee_coll_name'], k) for k, v in collection_info.items()])

# "Two way" dict to convert Earth Engine to/from geedim collection names
coll_names = dict(**gd_to_ee, **ee_to_gd)



"""
    Copyright 2021 Dugal Harris - dugalh@gmail.com

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

##
# Functionality for searching and compositing EE image collections
import collections
from datetime import datetime, timedelta

import ee
import geedim.image
import pandas as pd
from geedim import image, info, medoid


##
class Collection(object):
    def __init__(self, gd_coll_name):
        """
        Class for searching and compositing an EE image collection

        Parameters
        ----------
        gd_coll_name : str
                       geedim collection name:(landsat7_c2_l2|landsat8_c2_l2|sentinel2_toa|sentinel2_sr|modis_nbar)
        """
        if gd_coll_name not in info.gd_to_ee:
            raise ValueError(f"Unsupported collection: {gd_coll_name}")

        self._gd_coll_name = gd_coll_name
        self._ee_coll_name = info.gd_to_ee[self._gd_coll_name]
        self._collection_info = info.collection_info[gd_coll_name]
        self._image_class = image.get_class(gd_coll_name)  # geedim.image.*Image class for this collection
        self._ee_collection = None  # the wrapped ee.ImageCollection

        self._summary_key_df = pd.DataFrame(self._collection_info["properties"])  # key to metadata summary
        self._summary_df = None  # summary of the image metadata

    @classmethod
    def from_ids(cls, image_ids, mask=False):
        """
        Create collection from image IDs

        Parameters
        ----------
        image_ids : list(str)
                    A list of the EE image IDs (should all be from same collection)
        mask : bool, optional
               Apply a validity (cloud & shadow) mask to the image (default: False)

        Returns
        -------
        geedim.collection.Collection
        """
        # check image IDs are valid
        ee_coll_name = image.split_id(image_ids[0])[0]
        if ee_coll_name not in info.ee_to_gd:
            raise ValueError(f"Unsupported collection: {ee_coll_name}")

        id_check = [image.split_id(im_id)[0] == ee_coll_name for im_id in image_ids[1:]]
        if not all(id_check):
            raise ValueError("All images must belong to the same collection")

        # create the collection object
        gd_coll_name = info.ee_to_gd[ee_coll_name]
        gd_collection = cls(gd_coll_name)

        # build and wrap an ee.ImageCollection of processed (masked and scored) images
        im_list = ee.List([])
        for im_id in image_ids:
            gd_image = gd_collection._image_class.from_id(im_id, mask=mask)
            im_list = im_list.add(gd_image.ee_image)

        gd_collection._ee_collection = ee.ImageCollection(im_list)
        return gd_collection

    @property
    def ee_collection(self):
        """ee.ImageCollection : Returns the wrapped ee.ImageCollection"""
        return self._ee_collection

    @property
    def summary_key_df(self):
        """pandas.DataFrame : A key to Collection.summary_df
        (pandas.DataFrame with ABBREV and DESCRIPTION columns, and rows corresponding columns in summary_df)"""
        return self._summary_key_df

    @property
    def summary_df(self):
        """pandas.DataFrame : Summary of collection image properties with a row for each image"""
        if self._summary_df is None:
            self._summary_df = self._get_summary_df(self._ee_collection)
        return self._summary_df

    @property
    def summary_key(self):
        """str :  Formatted string of Collection.summary_key_df"""
        return self._summary_key_df[["ABBREV", "DESCRIPTION"]].to_string(index=False, justify="right")

    @property
    def summary(self):
        """str : Formatted string of Collection.summary_df"""
        return self.summary_df.to_string(
            float_format="%.2f",
            formatters={"DATE": lambda x: datetime.strftime(x, "%Y-%m-%d %H:%M")},
            columns=self._summary_key_df.ABBREV,
            index=False,
            justify="center",
        )

    composite_methods = ["q_mosaic", "mosaic", "median", "medoid"]  # supported composite methods

    def search(self, start_date, end_date, region, valid_portion=0):
        """
        Search for images based on date, region etc criteria

        Parameters
        ----------
        start_date : datetime.datetime
                     Start image capture date
        end_date : datetime.datetime
                   End image capture date (if None, then set to start_date + 1 day)
        region : dict, geojson, ee.Geometry
                 Polygon in WGS84 specifying a region that images should intersect
        valid_portion: int, optional
                       Minimum portion (%) of image pixels that should be valid (not cloud/shadow)

        Returns
        -------
        pandas.DataFrame
            Dataframe specifying image properties that match the search criteria
        """
        # Initialise
        if end_date is None:
            end_date = start_date + timedelta(days=1)
        if end_date <= start_date:
            raise ValueError("`end_date` must be at least a day later than `start_date`")

        def calc_stats(ee_image):
            """Server side calculation of validity and score stats within region of interest"""
            max_proj = geedim.image.get_projection(ee_image, min=False)
            gd_image = self._image_class(ee_image)
            region_sum = (
                ee.Image(1)
                .reduceRegion(reducer="sum", geometry=region, scale=max_proj.nominalScale(), crs=max_proj.crs())
            )

            # sum VALID_MASK and SCORE over image
            stats = (
                ee.Image([gd_image.masks["valid_mask"], gd_image.score])
                .unmask()
                .reduceRegion(reducer="sum", geometry=region, scale=max_proj.nominalScale(), crs=max_proj.crs())
                .rename(["VALID_MASK", "SCORE"], ["VALID_PORTION", "AVG_SCORE"])
            )

            # find average VALID_MASK and SCORE over region (not the same as image if image does not cover region)
            def region_mean(key, value):
                return ee.Number(value).divide(ee.Number(region_sum.get("constant")))

            stats = stats.map(region_mean)
            stats = stats.set("VALID_PORTION", ee.Number(stats.get("VALID_PORTION")).multiply(100))
            return gd_image.ee_image.set(stats)

        try:
            # filter the image collection, finding cloud/shadow masks, and region stats
            self._ee_collection = (
                self._image_class.ee_collection()
                .filterDate(start_date, end_date)
                .filterBounds(region)
                .map(calc_stats)
                .filter(ee.Filter.gt("VALID_PORTION", valid_portion))
            )
        finally:
            # update summary_df with image metadata from the filtered collection
            self._summary_df = self._get_summary_df(self._ee_collection)

        return self._summary_df

    def composite(self, method="q_mosaic", resampling='near'):
        """
        Create a cloud/shadow free composite.

        Note: composite() can be called on a filtered collection created by search(..), or on a collection created with
              fromIds(...)
              The `mask` parameter in search(...) and fromId(...) affects the composite and should generally be
              True so that cloud/shadow pixels are excluded.

        Parameters
        ----------
        method : str, optional
                 Compositing method to use (q_mosaic|mosaic|median|medoid).  (Default: q_mosaic).
        resampling : str, optional
               Resampling method for compositing and reprojecting: ("near"|"bilinear"|"bicubic") (default: "near")

        Returns
        -------
        : (ee.Image, str)
          The composite image, composite image ID
        """
        if resampling != 'near':
            self._ee_collection = self._ee_collection.map(lambda image: image.resample(resampling))

        method = str(method).lower()

        if method == "q_mosaic":
            comp_image = self._ee_collection.qualityMosaic("SCORE")
        elif method == "mosaic":
            comp_image = self._ee_collection.mosaic()
        elif method == "median":
            comp_image = self._ee_collection.median()
            # median creates float images, so re-apply any type conversion
            comp_image = self._image_class._im_transform(comp_image)
        elif method == "medoid":
            # limit medoid to surface reflectance bands
            sr_bands = [band_dict["id"] for band_dict in self._collection_info["bands"]]
            comp_image = medoid.medoid(self._ee_collection, bands=sr_bands)
        else:
            raise ValueError(f"Unsupported composite method: {method}")

        # populate image metadata with info on component images
        comp_image = comp_image.set("COMPONENT_IMAGES", self.summary)

        # construct an ID for the composite
        start_date = self.summary_df.DATE.iloc[0].strftime("%Y_%m_%d")
        end_date = self.summary_df.DATE.iloc[-1].strftime("%Y_%m_%d")

        comp_id = f"{self._ee_coll_name}/{start_date}-{end_date}-{method.upper()}_COMP"
        comp_image = comp_image.set("system:id", comp_id)
        # TODO: persist source CRS and scale by reprojecting?

        CompositeResult = collections.namedtuple("CompositeResult", ["image", "id"])
        return CompositeResult(comp_image, comp_id)

    def _get_summary_df(self, ee_collection):
        """
        Retrieve a summary of collection image metadata.

        Parameters
        ----------
        ee_collection : ee.ImageCollection
                        Filtered image collection whose image metadata to retrieve

        Returns
        -------
        : pandas.DataFrame
        pandas.DataFrame with a row of metadata for each image)
        """

        if ee_collection is None:
            return pd.DataFrame([], columns=self._summary_key_df.ABBREV)  # return empty dataframe

        # server side aggregation of relevant properties of ee_collection images
        init_list = ee.List([])

        def aggregrate_props(ee_image, prop_list):
            prop = ee.Dictionary()
            for prop_key in self._summary_key_df.PROPERTY.values:
                prop = prop.set(
                    prop_key, ee.Algorithms.If(ee_image.get(prop_key), ee_image.get(prop_key), ee.String("None"))
                )
            return ee.List(prop_list).add(prop)

        # retrieve list of dicts of collection image properties (the only call to getInfo() in Collection)
        im_prop_list = ee.List(ee_collection.iterate(aggregrate_props, init_list)).getInfo()

        if len(im_prop_list) == 0:
            return pd.DataFrame([], columns=self._summary_key_df.ABBREV)  # return empty dataframe

        # Convert ee.Date to python datetime
        for i, prop_dict in enumerate(im_prop_list):
            if "system:time_start" in prop_dict:
                prop_dict["system:time_start"] = datetime.utcfromtimestamp(prop_dict["system:time_start"] / 1000)

        # convert property list to DataFrame
        im_prop_df = pd.DataFrame(im_prop_list, columns=im_prop_list[0].keys())
        im_prop_df = im_prop_df.sort_values(by="system:time_start").reset_index(drop=True)  # sort by acquisition time
        im_prop_df = im_prop_df.rename(
            columns=dict(zip(self._summary_key_df.PROPERTY, self._summary_key_df.ABBREV))
        )  # abbreviate column names
        im_prop_df = im_prop_df[self._summary_key_df.ABBREV.to_list()]  # reorder columns

        return im_prop_df



"""
   Copyright 2021 Dugal Harris - dugalh@gmail.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

##
# Functions to download and export Earth Engine images

import os
import pathlib
import re
import time
import zipfile
from xml.dom import minidom
from xml.etree import ElementTree

import click
import ee
import requests

from geedim import image


def write_pam_xml(obj, filename):
    """
    Write image metadata to PAM xml file (as supported by GDAL/QGIS)

    Parameters
    ----------
    obj : ee.Image, geedim.image.Image, dict
          Image object whose metadata to write
    filename : str, pathlib.Path
               Path of the output file (use *.aux.xml)
    """
    # extract image metadata dict
    if isinstance(obj, dict):
        gd_info = obj
    elif isinstance(obj, ee.Image):
        gd_info = image.get_info(obj)
    elif isinstance(obj, image.Image):
        gd_info = obj.info
    else:
        raise TypeError(f"Unsupported type: {obj.__class__}")

    # remove footprint if it exists
    if "system:footprint" in gd_info["properties"]:
        gd_info["properties"].pop("system:footprint")

    # construct xml tree
    root = ElementTree.Element("PAMDataset")
    prop_meta = ElementTree.SubElement(root, "Metadata")
    for key, val in gd_info["properties"].items():
        item = ElementTree.SubElement(prop_meta, "MDI", attrib=dict(key=key))
        item.text = str(val)

    for band_i, band_dict in enumerate(gd_info["bands"]):
        band_elem = ElementTree.SubElement(root, "PAMRasterBand", attrib=dict(band=str(band_i + 1)))
        if "id" in band_dict:
            desc = ElementTree.SubElement(band_elem, "Description")
            desc.text = band_dict["id"]
        band_meta = ElementTree.SubElement(band_elem, "Metadata")
        for key, val in band_dict.items():
            item = ElementTree.SubElement(band_meta, "MDI", attrib=dict(key=key.upper()))
            item.text = str(val)

    # write indented xml string to file (excluding header)
    xml_str = minidom.parseString(ElementTree.tostring(root)).childNodes[0].toprettyxml(indent="   ")
    with open(filename, "w") as f:
        f.write(xml_str)


class _ExportImage(image.Image):
    """ Helper class for determining export/download crs, scale and region parameters"""
    def __init__(self, image_obj, name="Image", exp_region=None, exp_crs=None, exp_scale=None, resampling='near'):
        if isinstance(image_obj, image.Image):
            image.Image.__init__(self, image_obj.ee_image)
            self._info = image_obj.info
        else:
            image.Image.__init__(self, image_obj)

        self.name = name
        self.exp_region = exp_region
        self.exp_crs = exp_crs
        self.exp_scale = exp_scale
        self.resampling = resampling
        # TODO - what resampling to use and whether to expose CLI/API


    def parse_attributes(self):
        """ Set the exp_region, exp_crs and exp_scale attributes """

        if self.id is None:
            self._info["id"] = pathlib.Path(self.name).stem

        # if the image is in WGS84 and or has no scale (probable composite), then exit
        if ((self.scale is None) and (self.exp_scale is None)) or ((self.crs is None) and (self.exp_crs is None)):
            raise ValueError(f'{self.info["id"]} appears to be a composite in WGS84, specify a scale and CRS')

        # set resampling if image is not a composite
        if (self.scale is not None) and (self.crs is not None) and (self.resampling != 'near'):
            self._ee_image = self._ee_image.resample(self.resampling)

        # If CRS is the native MODIS CRS, then exit due to GEE bug
        if (self.crs == "SR-ORG:6974") and (self.exp_crs is None):
            raise ValueError(
                "There is an earth engine bug exporting in SR-ORG:6974, specify another CRS: "
                "https://issuetracker.google.com/issues/194561313"
            )

        if self.exp_crs is None:
            self.exp_crs = self.crs  # CRS corresponding to minimum scale band
        if self.exp_scale is None:
            self.exp_scale = self.scale  # minimum scale of the bands
        if self.exp_region is None:
            # use image granule footprint
            if "system:footprint" in self.info["properties"]:
                self.exp_region = self.info["properties"]["system:footprint"]
                click.secho(f'{self.info["id"]}: region not specified, setting to image footprint')
            else:
                raise AttributeError(f'{self.info["id"]} does not have a footprint, specify a region to download')

        if isinstance(self.exp_region, dict):
            self.exp_region = ee.Geometry(self.exp_region)


def export_image(image_obj, filename, folder="", region=None, crs=None, scale=None, resampling='near', wait=True):
    """
    Export an image to a GeoTiff in Google Drive

    Parameters
    ----------
    image_obj : ee.Image, geedim.image.Image
               The image to export
    filename : str
               The name of the task and destination file
    folder : str, optional
             Google Drive folder to export to (default: root).
    region : dict, geojson, ee.Geometry, optional
             Region of interest (WGS84) to export (default: export the entire image granule if it has one).
    crs : str, optional
          WKT, EPSG etc specification of CRS to export to.  Where image bands have different CRSs, all are re-projected
          to this CRS.
          (default: use the CRS of the minimum scale band if available).
    scale : float, optional
            Pixel scale (m) to export to.  Where image bands have different scales, all are re-projected to this scale.
            (default: use the minimum scale of image bands if available).
    resampling : str, optional
           Resampling method: ("near"|"bilinear"|"bicubic") (default: "near")
    wait : bool
           Wait for the export to complete before returning (default: True)

    Returns
    -------
    ee.batch.Task
        Earth Engine export task object
    """
    exp_image = _ExportImage(image_obj, name=filename, exp_region=region, exp_crs=crs, exp_scale=scale,
                             resampling=resampling)
    exp_image.parse_attributes()

    # create export task and start
    task = ee.batch.Export.image.toDrive(
        image=exp_image.ee_image,
        region=exp_image.exp_region,
        description=filename[:100],
        folder=folder,
        fileNamePrefix=filename,
        scale=exp_image.exp_scale,
        crs=exp_image.exp_crs,
        maxPixels=1e9,
    )

    task.start()

    if wait:  # wait for completion
        monitor_export_task(task)

    return task


def monitor_export_task(task, label=None):
    """
    Monitor and display the progress of an export task

    Parameters
    ----------
    task : ee.batch.Task
           WW task to monitor
    label: str, optional
           Optional label for progress display (default: use task description)
    """
    toggles = r"-\|/"
    toggle_count = 0
    pause = 0.5
    bar_len = 100
    status = ee.data.getOperation(task.name)

    if label is None:
        label = f'{status["metadata"]["description"][:80]}:'

    # wait for export preparation to complete, displaying a spin toggle
    while "progress" not in status["metadata"]:
        time.sleep(pause)
        status = ee.data.getOperation(task.name)  # get task status
        click.echo(f"\rPreparing {label} {toggles[toggle_count % 4]}", nl="")
        toggle_count += 1
    click.echo(f"\rPreparing {label}  done")

    # wait for export to complete, displaying a progress bar
    with click.progressbar(length=bar_len, label=f"Exporting {label}") as bar:
        while ("done" not in status) or (not status["done"]):
            time.sleep(pause)
            status = ee.data.getOperation(task.name)  # get task status
            progress = status["metadata"]["progress"] * bar_len
            bar.update(progress - bar.pos)  # update with progress increment
        bar.update(bar_len - bar.pos)

    if status["metadata"]["state"] != "SUCCEEDED":
        raise Exception(f"Export failed \n{status}")


def download_image(image_obj, filename, region=None, crs=None, scale=None, resampling='near', overwrite=False):
    """
    Download an image as a GeoTiff

    Parameters
    ----------
    image_obj : ee.Image, geedim.image.Image
               The image to export
    filename : str, pathlib.Path
               Name of the destination file
    region : dict, geojson, ee.Geometry, optional
             Region of interest (WGS84) to export (default: export the entire image granule if it has one).
    crs : str, optional
          WKT, EPSG etc specification of CRS to export to.  Where image bands have different CRSs, all are re-projected
          to this CRS.
          (default: use the CRS of the minimum scale band if available).
    scale : float, optional
            Pixel scale (m) to export to.  Where image bands have different scales, all are re-projected to this scale.
            (default: use the minimum scale of image bands if available).
    resampling : str, optional
           Resampling method: ("near"|"bilinear"|"bicubic") (default: "near")
    overwrite : bool, optional
                Overwrite the destination file if it exists, otherwise prompt the user (default: True)
    """
    filename = pathlib.Path(filename)
    exp_image = _ExportImage(image_obj, name=filename, exp_region=region, exp_crs=crs, exp_scale=scale,
                             resampling=resampling)
    exp_image.parse_attributes()

    # get download link
    try:
        link = exp_image.ee_image.getDownloadURL(dict(
            scale=exp_image.exp_scale,
            crs=exp_image.exp_crs,
            fileFormat="GeoTIFF",
            filePerBand=False,
            region=exp_image.exp_region,
        ))
    except ee.ee_exception.EEException as ex:
        # Add to exception message when the image is too large to download
        if re.match(r"Total request size \(.*\) must be less than or equal to .*", str(ex)):
            raise Exception(f"The requested image is too large to download, reduce its size, or export\n({str(ex)})")
        else:
            raise ex

    # download zip file
    tif_filename = filename.parent.joinpath(filename.stem + ".tif")  # force to tif file
    zip_filename = tif_filename.parent.joinpath("geedim_download.zip")
    with requests.get(link, stream=True) as r:
        r.raise_for_status()
        csize = 8192
        with open(zip_filename, "wb") as f:
            with click.progressbar(
                    r.iter_content(chunk_size=csize),
                    label=f"{filename.stem[:80]}:",
                    length=int(r.headers["Content-length"]) / csize,
                    show_pos=True,
            ) as bar:
                # override the bar's formatting function to show progress in MB
                bar.format_pos = (lambda: f"{bar.pos * csize / (2 ** 20):.1f}/{bar.length * csize / (2 ** 20):.1f} MB\r")
                for chunk in bar:
                    f.write(chunk)

    # extract image tif from zip file
    zip_tif_filename = zip_filename.parent.joinpath(zipfile.ZipFile(zip_filename, "r").namelist()[0])
    with zipfile.ZipFile(zip_filename, "r") as zip_file:
        zip_file.extractall(zip_filename.parent)
    os.remove(zip_filename)

    # rename to extracted tif file to specified filename
    if zip_tif_filename != tif_filename:
        while tif_filename.exists():
            if overwrite or click.confirm(f"{tif_filename.name} exists, do you want to overwrite?", default="n"):
                os.remove(tif_filename)
            else:
                tif_filename = click.prompt("Please enter another filename", type=str, default=None)
                tif_filename = pathlib.Path(tif_filename)
        os.rename(zip_tif_filename, tif_filename)

    # write image metadata to pam xml file
    write_pam_xml(exp_image.info, str(tif_filename) + ".aux.xml")

    return link
